# Grafové algoritmy - příklady
# Jan Kybic, 2017

import heapq

INF=float("inf") # nekonečno

# matice souslednosti (adjancency_matrix) pro orientovaný graf z přednášky o 6 vrcholech
g=[ [ INF for i in range(6) ] for j in range(6) ]
g[0][1]=5
g[0][5]=2
g[1][2]=4
g[2][3]=9
g[3][4]=7
g[3][5]=3
g[4][0]=1
g[5][2]=1
g[5][4]=8

def print_adjacency_matrix(g):
  for row in g:
    print(" ".join(map(lambda x: "%5.1f" % x,row)))

def all_distances_floyd(g):
  """ Floyd-Warshallův algoritmus - nalezne matici délek nejkratších cest mezi všemi dvojicemi uzlů
      na základě matice sousednosti 'g' s nezápornými hodnotami (cykly). """
  r=[ row.copy() for row in g ] # okopíruji matici g
  n=len(r)
  for k in range(n):   # cesta z uzlu do toho samého uzlu
    r[k][k]=0.        
  for k in range(n):
    for i in range(n):
      for j in range(n):
        r[i][j]=min(r[i][j],r[i][k]+r[k][j]) # cesta z 'i' do 'j' přes 'k'
  return r
     
def test_floyd():
  print_adjacency_matrix(g)
  r=all_distances_floyd(g)
  print("Floyd výsledky:")
  print_adjacency_matrix(r)

# reprezentace grafu pomoci seznamu sousednosti hran
# graf bude pole prvků typu Vertex
  
class Vertex:  
  def __init__(self):
    self.edges={}

  def addNeighbor(self,nbr,weight=0):
    self.edges[nbr]=weight   
    
# vytvoříme ten samý graf jako výše    
g2=[ Vertex() for i in range(6) ]
g2[0].addNeighbor(1,5)
g2[0].addNeighbor(5,2)
g2[1].addNeighbor(2,4)
g2[2].addNeighbor(3,9)
g2[3].addNeighbor(4,7)
g2[3].addNeighbor(5,3)
g2[4].addNeighbor(0,1)
g2[5].addNeighbor(2,1)
g2[5].addNeighbor(4,8)

def distances_dijkstra(g,start):
  """ Vstupem je graf 'g', reprezentovaný jako pole uzlů Vertex a index 'start' do tohoto pole.
      Najde nejkratší cesty z uzlu 'g[start]' do všech ostatních uzlů. Vrátí pole 'dist'
      obsahující tyto vzdálenosti a pole 'prev', obsahující indexy na uzel předcházející aktuálnímu na
      nejkratší cestě. """
  n=len(g)
  final=[False] * n   # je daný uzel definitivní?
  dist=[INF] * n      # zatím nejkratší vzdálenosti
  dist[start]=0
  prev=[None] * n     # pole předchůdců na optimální cestě  
  queue=[]            # prioritní fronta, srovnaná dle vzdálenosti - (vzdálenost, index uzlu)
  heapq.heappush(queue,(0.,start)) # vložíme počáteční uzel do fronty
  while len(queue)>0:
    i=heapq.heappop(queue)[1]    # vrchol s nejmenší vzdáleností
    if final[i]:                 # už byl definitivní, ignorujeme  
      continue
    final[i]=True                # teď už je definitivní, kratší cesta není
    for j,w in g[i].edges.items():  # projde všechny sousedy 'i'
      altd=dist[i]+w                   # délka alternativní cesty do 'j'
      if altd<dist[j]:                 # alternativní cesta je kratší
        dist[j]=altd
        prev[j]=i
        heapq.heappush(queue,(altd,j))
  return dist,prev

def dijkstra_path(prev,i):
  """ Vstupem je pole 'prev' vypočítané funkcí 'distances_dijkstra'. Pro vrchol 'i' vrátí posloupnost
      vrcholů na nejkratší cestě ze 'start' do 'i'. Totéž jako podobné dfs2.find_path() """
  path=[]
  while i is not None:
    path+=[i]
    i=prev[i]
  return list(reversed(path))  

def test_dijkstra():
   dist,prev=distances_dijkstra(g2,0)
   print("Nejkratší vzdálenosti z uzlu 0:", dist)
   for i in range(len(g)):
     print("Nejkratší cesta z uzlu 0 do uzlu %d:" % i, dijkstra_path(prev,i))


if __name__=="__main__":
  test_floyd()  
  test_dijsktra()
